/*
 * yasapi_config2.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#if defined (YASAPI_CONFIG2) && ! defined (YASAPI_CONFIG3) // {
#include <uxtheme.h>
#include <resources.h>

///////////////////////////////////////////////////////////////////////////////
typedef struct _PageVMT PageVMT;
typedef struct _PageTemplate PageTemplate;
typedef struct _Page Page;
typedef struct _Priv Priv;

///////////////////////////////////////////////////////////////////////////////
static void SetDeviceOptions(Priv *pPriv, const Options *pOptions);

///////////////////////////////////////////////////////////////////////////////
struct _PageVMT {
  const char *name;
  DLGPROC lpDialogFunc;
  void (*SetDeviceOptions)(HWND hDlg, const Options *pOptions, int nDevice);
  void (*GetDeviceOptions)(HWND hDlg, Options *pOptions, int nDevice);
};

struct _PageTemplate {
  char *label;
  int idd;
  PageVMT *(*GetVMT)(void);
};

struct _Page {
  PageVMT *vmt;
  HWND hDlg;
};

struct _Priv {
  Options *pOptions;
  HMODULE hModule;
  HWND hDlgPage;
  int nDevice;
  int nNumPages;
  int nPage;
  PageTemplate *pTemplates;
  Page aPages[0];
};

///////////////////////////////////////////////////////////////////////////////
static PageVMT *GetDevicesVMT(void);
static PageVMT *GetOptionsVMT(void);
static PageVMT *GetBuffersVMT(void);

static PageTemplate gaTemplates[]={
  { "Devices",IDD_PAGE_DEVICES,GetDevicesVMT },
  { "Device Options",IDD_PAGE_OPTIONS,GetOptionsVMT },
  { "Device Buffers",IDD_PAGE_BUFFERS,GetBuffersVMT }
};

///////////////////////////////////////////////////////////////////////////////
static void DevicesFillList(HWND hDlg, Priv *pPriv, int idc)
{
  Options *pOptions=pPriv->pOptions;
  DWORD cDevices=pOptions->cDevices,nDevice;
  DeviceOptions *pDevices=pOptions->pDevices;
#if defined (YASAPI_DEBUG) // {
  int debug=pOptions->debug;
#endif // }
  HWND hWndList;

  if (NULL==(hWndList=GetDlgItem(hDlg,idc))) {
    DMESSAGE(debug,"getting list");
    goto list;
  }

  for (nDevice=0;nDevice<cDevices;++nDevice) {
    SendMessageW(hWndList,LB_ADDSTRING,0,
        (LPARAM)pDevices[nDevice].val_Device_FriendlyName.pwszVal);
  }

  SendMessageW(hWndList,LB_SETCURSEL,pPriv->nDevice,0);
list:
  return;
}

static void DevicesOnSelChange(HWND hDlg, Priv *pPriv, HWND hWndList)
{
  pPriv->nDevice=SendMessageW(hWndList,LB_GETCURSEL,0,0);
  SetDeviceOptions(pPriv,pPriv->pOptions);
}

static INT_PTR CALLBACK DevicesProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  Priv *pPriv=(Priv *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch (uMsg) {
  case WM_INITDIALOG:
    pPriv=(Priv *)lParam;
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    DevicesFillList(hDlg,pPriv,IDC_LIST_DEVICES);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDC_LIST_DEVICES:
      switch (HIWORD(wParam)) {
      case LBN_SELCHANGE:
        DevicesOnSelChange(hDlg,pPriv,(HWND)lParam);
        return TRUE;
      default:
        break;
      }

      break;
    default:
      break;
    }

    break;
  default:
    break;
  }

  return FALSE;
}

static void DevicesSetDeviceOptions(HWND hDlg, const Options *pOptions,
    int nDevice)
{
}

static void DevicesGetDeviceOptions(HWND hDlg, Options *pOptions, int nDevice)
{
  HWND hWndList;

  if (NULL==(hWndList=GetDlgItem(hDlg,IDC_LIST_DEVICES))) {
    DMESSAGE(pOptions->debug,"getting handle of listbox");
    goto list;
  }

  SendMessageW(hWndList,LB_GETSELITEMS,1,(LPARAM)&pOptions->nDevice);
list:
  return;
}

static PageVMT *GetDevicesVMT(void)
{
  static PageVMT vmt;

  if (NULL==vmt.name) {
    vmt.name="Devices";
    vmt.lpDialogFunc=DevicesProc;
    vmt.SetDeviceOptions=DevicesSetDeviceOptions;
    vmt.GetDeviceOptions=DevicesGetDeviceOptions;
  }

  return &vmt;
}

///////////////////////////////////////////////////////////////////////////////
static INT_PTR CALLBACK OptionsProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  Priv *pPriv=(Priv *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch (uMsg) {
  case WM_INITDIALOG:
    pPriv=(Priv *)lParam;
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDC_RADIO_EXCLUSIVE:
    case IDC_RADIO_SHARED:
    case IDC_RADIO_AUTOMATIC:
    case IDC_CHECKBOX_AUTOCONVERT_PCM:
    case IDC_CHECKBOX_SRC_DEFAULT_QUALITY:
      ConfigOptionsSyncShareMode(hDlg);
      return TRUE;
    default:
      break;
    }

    break;
  default:
    break;
  }

  return FALSE;
}

static PageVMT *GetOptionsVMT(void)
{
  static PageVMT vmt;

  if (NULL==vmt.name) {
    vmt.name="Options";
    vmt.lpDialogFunc=OptionsProc;
    vmt.SetDeviceOptions=ConfigOptionsSetDeviceOptions;
    vmt.GetDeviceOptions=ConfigOptionsGetDeviceOptions;
  }

  return &vmt;
}

///////////////////////////////////////////////////////////////////////////////
static INT_PTR CALLBACK BuffersProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  Priv *pPriv=(Priv *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch (uMsg) {
  case WM_INITDIALOG:
    pPriv=(Priv *)lParam;
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    return TRUE;
  case WM_HSCROLL:
    ConfigSliderSync(hDlg);
    break;
  default:
    break;
  }

  return FALSE;
}

static PageVMT *GetBuffersVMT(void)
{
  static PageVMT vmt;

  if (NULL==vmt.name) {
    vmt.name="Buffers";
    vmt.lpDialogFunc=BuffersProc;
    vmt.SetDeviceOptions=ConfigOptionsSetDeviceBuffers;
    vmt.GetDeviceOptions=ConfigOptionsGetDeviceBuffers;
  }

  return &vmt;
}

///////////////////////////////////////////////////////////////////////////////
static void OnSelChange(HWND hDlg, Priv *pPriv, HWND hWndTab)
{
  int nPage=TabCtrl_GetCurSel(hWndTab);
  Page *pPage=pPriv->aPages+nPage;

  if (pPriv->hDlgPage) {
    ShowWindow(pPriv->hDlgPage,FALSE);
    pPriv->hDlgPage=NULL;
  }

  pPriv->nPage=nPage;
  pPriv->hDlgPage=pPage->hDlg;
  ShowWindow(pPriv->hDlgPage,TRUE);
}

static int InitPage(HWND hDlg, HWND hWndTab, Priv *pPriv, int nPage)
{
  int code=-1;
  PageTemplate *pTemplate=pPriv->pTemplates+nPage;
  Page *pPage=pPriv->aPages+nPage;
#if defined (YASAPI_DEBUG) // {
  int debug=pPriv->pOptions->debug;
#endif // }
  TCITEMA tie;
  LPCTSTR lpTemplateName;
  RECT rc;

  pPage->vmt=pTemplate->GetVMT();

  memset(&tie,0,sizeof tie);
  tie.mask=TCIF_TEXT|TCIF_IMAGE;
  tie.iImage=-1;
  tie.pszText=pTemplate->label;
  SendMessageA(hWndTab,TCM_INSERTITEMA,nPage,(LPARAM)&tie);

  // Finding resource.
  lpTemplateName=MAKEINTRESOURCEA(pTemplate->idd);

  pPage->hDlg=CreateDialogParamA(
    pPriv->hModule,   // _In_opt_ HINSTANCE hInstance,
    lpTemplateName,   // _In_     LPCTSTR   lpTemplateName,
    hDlg,             // _In_opt_ HWND      hWndParent,
    pPage->vmt->lpDialogFunc,
                      // _In_opt_ DLGPROC   lpDialogFunc,
    (LPARAM)pPriv     // _In_     LPARAM    dwInitParam
  );

  if (NULL==pPage->hDlg) {
    DMESSAGE(debug,"creating child dialog");
    goto child;
  }

  GetClientRect(pPage->hDlg,&rc);
  TabCtrl_AdjustRect(hWndTab,FALSE,&rc);

  MapWindowPoints(
    hWndTab,          // _In_    HWND    hWndFrom,
    hDlg,             // _In_    HWND    hWndTo,
    (LPPOINT)&rc,     // _Inout_ LPPOINT lpPoints,
    2                 // _In_    UINT    cPoints
  );

  SetWindowPos(pPage->hDlg,HWND_TOP,rc.left,rc.top,0,0,
      SWP_NOSIZE|SWP_HIDEWINDOW);
  code=0;
child:
lock:
load:
find:
  return code;
}

static int InitTab(HWND hDlg, int idc, Priv *pPriv)
{
  int code=-1;
#if defined (YASAPI_DEBUG) // {
  int debug=pPriv->pOptions->debug;
#endif // }
  HWND hWndTab;
  int nPage;

  // Get the tabcontrol handle.
  if (NULL==(hWndTab=GetDlgItem(hDlg,idc))) {
    DMESSAGE(debug,"getting tabcontrol handle");
    goto item;
  }

  // Initialize pages.
  for (nPage=0;nPage<pPriv->nNumPages;++nPage) {
    if (InitPage(hDlg,hWndTab,pPriv,nPage)<0) {
      DMESSAGE(debug,"initializing page");
      goto param;
    }

#if 1 // {
    EnableThemeDialogTexture(
      pPriv->aPages[nPage].hDlg,  // _In_ HWND  hwnd,
      ETDT_ENABLETAB              // _In_ DWORD dwFlags
    );
#endif // }
  }

  SendMessageA(hWndTab,TCM_SETCURSEL,pPriv->nPage,0);
  OnSelChange(hDlg,pPriv,hWndTab);
  code=0;
param:
item:
  return code;
}

static void SetDeviceOptions(Priv *pPriv, const Options *pOptions)
{
  int nPage;
  Page *pPage;

  for (nPage=0;nPage<pPriv->nNumPages;++nPage) {
    pPage=pPriv->aPages+nPage;
    pPage->vmt->SetDeviceOptions(pPage->hDlg,pOptions,pPriv->nDevice);
  }
}

static void SetOptions(HWND hDlg, Priv *pPriv, const Options *pOptions)
{
  ConfigOptionsSetOptions(hDlg,pOptions);
  SetDeviceOptions(pPriv,pOptions);
}

static void GetDeviceOptions(Priv *pPriv, Options *pOptions)
{
  int nPage;
  Page *pPage;

  for (nPage=0;nPage<pPriv->nNumPages;++nPage) {
    pPage=pPriv->aPages+nPage;
    pPage->vmt->GetDeviceOptions(pPage->hDlg,pOptions,pPriv->nDevice);
  }
}

static void GetOptions(HWND hDlg, Priv *pPriv, Options *pOptions)
{
  GetDeviceOptions(pPriv,pOptions);
  ConfigOptionsGetOptions(hDlg,pOptions);
}

static INT_PTR CALLBACK ConfigProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  Priv *pPriv=(Priv *)GetWindowLongPtr(hDlg,GWLP_USERDATA);
#if defined (YASAPI_DEBUG) // {
  int debug;
#endif // }

  switch (uMsg) {
  case WM_INITDIALOG:
    pPriv=(Priv *)lParam;
#if defined (YASAPI_DEBUG) // {
    debug=pPriv->pOptions->debug;
#endif // }
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    SetWindowTextA(hDlg,PI_LABEL " Settings");
#if ! defined (YASAPI_OPTION_WRITTEN_TIME) // {
    ShowWindow(GetDlgItem(hDlg,IDC_STATIC_LABEL_WRITTENTIME),FALSE);
    ShowWindow(GetDlgItem(hDlg,IDC_CHECKBOX_WRITTENTIME_AUDIO),FALSE);
    ShowWindow(GetDlgItem(hDlg,IDC_CHECKBOX_WRITTENTIME_VIDEO),FALSE);
#endif // }

    if (InitTab(hDlg,IDC_TABCONTROL,pPriv)<0) {
      DMESSAGE(debug,"initializing tabcontrol");
      break;
    }

    SetOptions(hDlg,pPriv,pPriv->pOptions);
    break;
  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDOK:
      pPriv->pOptions->nPage=pPriv->nPage;
      OptionsSetDevice(pPriv->pOptions,pPriv->nDevice);
      GetOptions(hDlg,pPriv,pPriv->pOptions);
      EndDialog(hDlg,TRUE);
      return TRUE;
    case IDCANCEL:
      EndDialog(hDlg,TRUE);
      return TRUE;
    case IDC_BUTTON_DEFAULT:
      SetOptions(hDlg,pPriv,NULL);
      return TRUE;
    default:
      break;
    }

    break;
  case WM_NOTIFY:
    switch (((NMHDR *)lParam)->idFrom) {
    case IDC_TABCONTROL:
      switch (((NMHDR *)lParam)->code) {
      case TCN_SELCHANGE:
        OnSelChange(hDlg,pPriv,((NMHDR *)lParam)->hwndFrom);
        return TRUE;
      default:
        break;
      }

      break;
    default:
      break;
    }

    break;
  default:
    break;
  }

  return FALSE;
}

void ConfigDialog(HWND hWndParent, Options *pOptions)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pOptions->debug;
#endif // }
  int nNumPages,nPage;
  size_t size;
  Priv *pPriv;
  LPCSTR lpTemplateName;

  nNumPages=(sizeof gaTemplates)/(sizeof *gaTemplates);
  size=(sizeof *pPriv)+nNumPages*(sizeof *pPriv->aPages);

  if (NULL==(pPriv=malloc(size))) {
    DMESSAGE(debug,"allocating priv");
    goto priv;
  }

  memset(pPriv,0,size);
  pPriv->pOptions=pOptions;
  pPriv->hModule=GetModuleHandleA(YASAPI_MODULE);
  pPriv->hDlgPage=NULL;
  pPriv->nDevice=pOptions->nDevice;
  pPriv->nNumPages=nNumPages;
  pPriv->nPage=pOptions->nPage<nNumPages?pOptions->nPage:0;
  pPriv->pTemplates=gaTemplates;

  lpTemplateName=MAKEINTRESOURCEA(IDD_CONFIG2);

  DialogBoxParamA(
    pPriv->hModule,       // _In_opt_  HINSTANCE hInstance,
    lpTemplateName,       // _In_      LPCTSTR lpTemplateName,
    hWndParent,           // _In_opt_  HWND hWndParent,
    ConfigProc,           // _In_opt_  DLGPROC lpDialogFunc,
    (LPARAM)pPriv         // _In_      LPARAM dwInitParam
  );

  free(pPriv);
priv:
  return;
}
#endif // }
