/*
 * yasapi_device_options.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#if defined (YASAPI_CONFIG2) && ! defined (PKEY_Device_FriendlyName) // {
#include <FunctionDiscoveryKeys_devpkey.h>
#endif // }

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_KEY_SHARE_MODE           L"share_mode"
#define YASAPI_KEY_PULL                 L"pull"
#define YASAPI_KEY_AUTOCONVERTPCM       L"autoconvertpcm"
#define YASAPI_KEY_SRC_DEFAULT_QUALITY  L"src_default_quality"
#define YASAPI_KEY_RING_SIZE            L"ring_size"
#define YASAPI_KEY_RING_FILL            L"ring_fill"
#define YASAPI_KEY_SHARE_SIZE           L"share_size"

#if defined (YASAPI_CONFIG2) // {
///////////////////////////////////////////////////////////////////////////////
int DeviceOptionsInit(DeviceOptions *pOptions, Plugin *pPlugin, int nDevice)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  IMMDeviceCollection *pCollection=pPlugin->pCollection;
  IMMDevice *pDevice;
  LPWSTR pstrId;
  IPropertyStore *pProperties;
  HRESULT hr;

  // Copy the path.
  pOptions->path=pPlugin->pOptions->path;

  // get a pointer to the specified item in the device collection.
  hr=pCollection->lpVtbl->Item(pCollection,
    nDevice,                // [in]  UINT      nDevice,
    &pDevice                // [out] IMMDevice **ppDevice
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DMESSAGE(debug,"getting item from device collection");
    goto device;
  }

  DPUTS(debug,"    got item from device collection\n");
  pOptions->pDevice=pDevice;

  // retrieving an endpoint ID string that identifies the
  // audio endpoint device.
  hr=pDevice->lpVtbl->GetId(pDevice,
    &pstrId                 // [out] LPWSTR *ppstrId
  );

  if (FAILED(hr)) {
    DERROR(debug,E_OUTOFMEMORY,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"retrieving endpoint ID string");
    goto strid;
  }

  wcsncpy(pOptions->id,pstrId,YASAPI_ID_SIZE-1);
  pOptions->id[YASAPI_ID_SIZE-1]=L'\0';
  CoTaskMemFree(pstrId);
  DWPRINTF(debug,L"    id: \"%s\"\n",pOptions->id);

  // retrieving an interface to the device's property store.
  hr=pDevice->lpVtbl->OpenPropertyStore(pDevice,
    STGM_READ,              // [in]  DWORD          stgmAccess,
    &pProperties            // [out] IPropertyStore **ppProperties
  );

  if (FAILED(hr)) {
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGE(debug,"retrieving interface to property store");
    goto properties;
  }

  DPUTS(debug,"    got interface to property store\n");
  pOptions->pProperties=pProperties;

  // getting the friendly name of the audio adapter to which the
  // endpoint device is attached
  PropVariantInit(&pOptions->val_Device_FriendlyName);

  hr=pProperties->lpVtbl->GetValue(pProperties,
    &PKEY_Device_FriendlyName,
                            // [in]  REFPROPERTYKEY key,
    &pOptions->val_Device_FriendlyName
                            // [out] PROPVARIANT    *pv
  );

  if (FAILED(hr)) {
    DMESSAGE(debug,"retrieving PKEY_Device_FriendlyName property");
    goto friendly;
  }

  if (VT_LPWSTR==pOptions->val_Device_FriendlyName.vt) {
    DWPRINTF(debug,L"    PKEY_Device_FriendlyName: \"%s\"\n",
        pOptions->val_Device_FriendlyName.pwszVal);
  }

  // initializing hnsDevicePeriod.
  pOptions->hnsDevicePeriod=0;

  // loading device options.
  DeviceOptionsLoad(pOptions,pOptions->id,GetDefaultDeviceOptions);

  return 0;
// cleanup:
  //PropVariantClear(&pOptions->val_Device_FriendlyName);
friendly:
  pProperties->lpVtbl->Release(pProperties);
properties:
strid:
  pDevice->lpVtbl->Release(pDevice);
device:
  return -1;
}

void DeviceOptionsCleanup(DeviceOptions *pOptions)
{
  PropVariantClear(&pOptions->val_Device_FriendlyName);
  pOptions->pProperties->lpVtbl->Release(pOptions->pProperties);
  pOptions->pDevice->lpVtbl->Release(pOptions->pDevice);
}

const DeviceOptions *GetDefaultDeviceOptions(void)
{
  static int initialized=0;
  static DeviceOptions options;

  if (0==initialized) {
#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
    options.eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
#else // } {
    options.mode=YASAPI_MODE_EXCLUSIVE;
#endif // }
    options.pull=FALSE;
    options.autoconvertpcm=TRUE;
    options.src_default_quality=FALSE;
    options.qRingSize=4.0;
    options.qRingFill=2.5;
    options.qShareSize=0.0;
    initialized=1;
  }

  return &options;
}
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_CONFIG2) // {
void DeviceOptionsLoad(DeviceOptions *pOptions, wchar_t *app,
    const DeviceOptions *(*GetDefaultOptions)(void))
#else // } {
void DeviceOptionsLoad(Options *pOptions, wchar_t *app,
    const Options *(*GetDefaultOptions)(void))
#endif // }
{
  enum { SIZE=32 };
  wchar_t set[SIZE];
  wchar_t get[SIZE];

  // get the share mode.
#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
  pOptions->eShareMode=GetPrivateProfileIntW(
    app,                          // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_SHARE_MODE,        // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->eShareMode,
                                  // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );
#else // } {
  pOptions->mode=GetPrivateProfileIntW(
    app,                          // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_SHARE_MODE,        // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->mode,    // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );
#endif // }

  // get the pull flag.
  pOptions->pull=GetPrivateProfileIntW(
    app,                          // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_PULL,              // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->pull,    // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // get the autoconvert pcm flag.
  pOptions->autoconvertpcm=GetPrivateProfileIntW(
    app,                          // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_AUTOCONVERTPCM,    // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->autoconvertpcm,
                                  // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // get the src default quality flag.
  pOptions->src_default_quality=GetPrivateProfileIntW(
    app,                          // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_SRC_DEFAULT_QUALITY,
                                  // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->src_default_quality,
                                  // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // get the ring size.
  swprintf(set,SIZE,L"%f",GetDefaultOptions()->qRingSize);

  GetPrivateProfileStringW(
    app,                          // _In_   LPCTSTR lpAppName,
    YASAPI_KEY_RING_SIZE,         // _In_   LPCTSTR lpKeyName,
    set,                          // _In_   LPCTSTR lpDefault,
    get,                          // _Out_  LPTSTR lpReturnedString,
    SIZE,                         // _In_   DWORD nSize,
    pOptions->path                // _In_   LPCTSTR lpFileName
  );

  pOptions->qRingSize=_wtof(get);

  // get the ring fill.
  swprintf(set,SIZE,L"%f",GetDefaultOptions()->qRingFill);

  GetPrivateProfileStringW(
    app,                          // _In_   LPCTSTR lpAppName,
    YASAPI_KEY_RING_FILL,         // _In_   LPCTSTR lpKeyName,
    set,                          // _In_   LPCTSTR lpDefault,
    get,                          // _Out_  LPTSTR lpReturnedString,
    SIZE,                         // _In_   DWORD nSize,
    pOptions->path                // _In_   LPCTSTR lpFileName
  );

  pOptions->qRingFill=_wtof(get);

  // get the share size.
  swprintf(set,SIZE,L"%f",GetDefaultOptions()->qShareSize);

  GetPrivateProfileStringW(
    app,                          // _In_   LPCTSTR lpAppName,
    YASAPI_KEY_SHARE_SIZE,        // _In_   LPCTSTR lpKeyName,
    set,                          // _In_   LPCTSTR lpDefault,
    get,                          // _Out_  LPTSTR lpReturnedString,
    SIZE,                         // _In_   DWORD nSize,
    pOptions->path                // _In_   LPCTSTR lpFileName
  );

  pOptions->qShareSize=_wtof(get);
}

#if defined (YASAPI_CONFIG2) // {
void DeviceOptionsSave(DeviceOptions *pOptions, wchar_t *app)
#else // } {
void DeviceOptionsSave(Options *pOptions, wchar_t *app)
#endif // }
{
  enum { SIZE=32 };
  wchar_t buf[SIZE];

  // clear hnsDevicePeriod.
  pOptions->hnsDevicePeriod=0;

  // write the share size.
  swprintf(buf,SIZE,L"%f",pOptions->qShareSize);

  WritePrivateProfileStringW(
    app,                          // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_SHARE_SIZE,        // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the ring fill.
  swprintf(buf,SIZE,L"%f",pOptions->qRingFill);

  WritePrivateProfileStringW(
    app,                          // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_RING_FILL,         // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the ring size.
  swprintf(buf,SIZE,L"%f",pOptions->qRingSize);

  WritePrivateProfileStringW(
    app,                          // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_RING_SIZE,         // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the src default quality flag.
  swprintf(buf,SIZE,L"%d",pOptions->src_default_quality);

  WritePrivateProfileStringW(
    app,                          // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_SRC_DEFAULT_QUALITY,
                                  // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the autoconvert pcm flag.
  swprintf(buf,SIZE,L"%d",pOptions->autoconvertpcm);

  WritePrivateProfileStringW(
    app,                          // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_AUTOCONVERTPCM,    // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the pull flag.
  swprintf(buf,SIZE,L"%d",pOptions->pull);

  WritePrivateProfileStringW(
    app,                          // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_PULL,              // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the share mode.
#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
  swprintf(buf,SIZE,L"%d",pOptions->eShareMode);
#else // } {
  swprintf(buf,SIZE,L"%d",pOptions->mode);
#endif // }

  WritePrivateProfileStringW(
    app,                          // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_SHARE_MODE,        // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );
}
