/*
 * yasapi_player.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#include <avrt.h>

///////////////////////////////////////////////////////////////////////////////
// Rendering a Stream
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd316756%28v=vs.85%29.aspx
//
// Exclusive-Mode Streams
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd370844%28v=vs.85%29.aspx

///////////////////////////////////////////////////////////////////////////////
#if ! defined (AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY) // {
#define AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY 0x08000000
#endif // }

#if ! defined(AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM) // {
#define AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM 0x80000000
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if 0 // {
#define PLAYER_FRAMES_MS(nFrames,pwfx) \
  MulDiv(nFrames,1000,(pwfx)->nSamplesPerSec)
#define PLAYER_FRAMES_TIMER(nFrames,pwfx) \
  (10000000.0*(nFrames)/(pwfx)->nSamplesPerSec)
#else // } {
#define PLAYER_FRAMES_MS(nFrames,pwfx) \
  MulDiv(1000,nFrames,(pwfx)->nSamplesPerSec)
#define PLAYER_FRAMES_TIMER(nFrames,pwfx) \
  (10000.0*1000*(nFrames)/(pwfx)->nSamplesPerSec)
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_DEBUG) // {
//#define YASAPI_PLAYER_GET_MIX_FORMAT
//#define YASAPI_PLAYER_IS_FORMAT_SUPPORTED
#endif // }
//#define YASAPI_PLAYER_GET_STREAM_LATENCY

static IAudioClient *PlayerActivate(Player *pPlayer)
{
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_CONFIG2) // {
  Options *pOptions=pPlugin->pOptions;
  IMMDevice *pDevice=pOptions->pDevices[pOptions->nDevice].pDevice;
#else // } {
  IMMDevice *pDevice=pPlugin->pDevice;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  HRESULT hr;

  // 2. Call IAudioClient::Release to release the audio client used in the
  //    previous call that returned AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED.
  if (NULL!=pAudioClient) {
    pAudioClient->lpVtbl->Release(pAudioClient);
    DPUTS(debug,"  audio client destroyed\n");
    pPlayer->pAudioClient=NULL;
    pAudioClient=NULL;
  }

  // 4. Call the IMMDevice::Activate method with parameter iid set to
  //    REFIID IID_IAudioClient to create a new audio client.
  hr=pDevice->lpVtbl->Activate(pDevice,
    &IID_IAudioClient,        // [in]   REFIID iid,
    CLSCTX_ALL,               // [in]   DWORD dwClsCtx,
    NULL,                     // [in]   PROPVARIANT *pActivationParams,
    (void **)&pAudioClient    // [out]  void **ppInterface
  );

  if (FAILED(hr)) {
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DMESSAGE(debug,"activating audio client");
    goto client;
  }

  pPlayer->pAudioClient=pAudioClient;
  DPUTS(debug,"  audio client activated\n");

  return pAudioClient;
client:
  return NULL;
}

int PlayerCreate(Player *pPlayer, Plugin *pPlugin, int samplerate,
    int numchannels, int bitspersamp)
{
  Options *pOptions=pPlugin->pOptions;
#if defined (YASAPI_DEBUG) // {
  int debug=pOptions->debug;
#else // } {
  int debug=0;
#endif // }
#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
  AUDCLNT_SHAREMODE eShareMode=OptionsGetShareMode(pOptions);
#else // }
  int mode=OptionsGetMode(pOptions);
#endif // }
  int pull=OptionsGetPull(pOptions);
  int autoconvert_pcm=OptionsGetAutoConvertPCM(pOptions);
  int src_default_quality=OptionsGetSRCDefaultQuality(pOptions);
  double qRingSize=OptionsGetRingSize(pOptions);
  double qShareSize=OptionsGetShareSize(pOptions);
  REFERENCE_TIME hnsDevicePeriod=OptionsGetDevicePeriod(pOptions);
  WAVEFORMATEXTENSIBLE *pwfxx=&pPlayer->wfxx;
  WAVEFORMATEX *pwfx=&pwfxx->Format;
  int retry=0;
#if defined (YASAPI_CONFIG2) // {
  IMMDevice *pDevice=pOptions->pDevices[pOptions->nDevice].pDevice;
#else // } {
  IMMDevice *pDevice=pPlugin->pDevice;
#endif // }
#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
  DWORD dwStreamFlags=0;
#else // {
  AUDCLNT_SHAREMODE eShareMode;
  DWORD dwStreamFlags;
#endif // }
  IAudioClient *pAudioClient;
  HRESULT hr;
  UINT32 nBlockAlign;
  IAudioRenderClient *pRenderClient;
  IAudioClock *pAudioClock;
#if defined (YASAPI_PLAYER_GET_MIX_FORMAT) // {
  WAVEFORMATEX *pDeviceFormat;
#endif // }
#if defined (YASAPI_PLAYER_IS_FORMAT_SUPPORTED) // {
  WAVEFORMATEX *pClosestMatch,**ppClosestMatch;
#endif // }
#if defined (YASAPI_PLAYER_GET_STREAM_LATENCY) // {
  REFERENCE_TIME hnsLatency;
#endif // }
  UINT32 uNumBufferFrames,uBufSize;
  SyncConfig sc;

  if (0==pPlugin->initialized) {
    DMESSAGE(debug,"not initialized");
    goto initialized;
  }

  // initialize the the strategy.
  pPlayer->strategy=pull?PlayerPull():PlayerPush();
  DPRINTF(debug,"  initializing for %s mode\n",pPlayer->strategy->name);
  pPlayer->pPlugin=pPlugin;

  // setup WAVEFORMATEXTENSIBLE.
  if (1==numchannels&&0!=pOptions->mono2stereo) {
    pPlayer->mono=1;
    numchannels=2;
  }
  else
    pPlayer->mono=0;

  WFXXSetup(pwfxx,samplerate,numchannels,bitspersamp,FALSE,FALSE);
  pPlayer->wBytesPerSample=pwfx->wBitsPerSample>>3;

  // initialize the audio client.
  pPlayer->pAudioClient=NULL;

  if (NULL==(pAudioClient=PlayerActivate(pPlayer)))
    goto client;

#if defined (YASAPI_PLAYER_GET_MIX_FORMAT) // {
  if (0!=debug) {
    // retrieve the stream format that the audio engine uses for its
    // internal processing of shared-mode streams.
    hr=pAudioClient->lpVtbl->GetMixFormat(pAudioClient,
      &pDeviceFormat          // [out]  WAVEFORMATEX **ppDeviceFormat
    );

    if (FAILED(hr)) {
      DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
      DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
      DERROR(debug,E_POINTER,hr);
      DERROR(debug,E_OUTOFMEMORY,hr);
      DMESSAGE(debug,"getting device format");
      goto format;
    }
    else if (NULL!=pDeviceFormat) {
      if (debug) {
        fputs("  internal shared mode format: ",stderr);
        WFXPrint(pDeviceFormat,stderr);
        fputs("\n",stderr);
      }

      CoTaskMemFree(pDeviceFormat);
    }
  }
#endif // }

#if ! defined (YASAPI_AUDCLNT_SHAREMODE) // {
  switch (mode) {
  case YASAPI_MODE_EXCLUSIVE:
  case YASAPI_MODE_AUTOMATIC:
    eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
    break;
  default:
    eShareMode=AUDCLNT_SHAREMODE_SHARED;
    break;
  }
#endif // }

#if defined (YASAPI_PLAYER_IS_FORMAT_SUPPORTED) // {
  // check whether the format is supported.
  pClosestMatch=NULL;
  ppClosestMatch=AUDCLNT_SHAREMODE_EXCLUSIVE==eShareMode?NULL:&pClosestMatch;

  hr=pAudioClient->lpVtbl->IsFormatSupported(pAudioClient,
    eShareMode,             // [in]   AUDCLNT_SHAREMODE ShareMode,
    pwfx,                   // [in]   const WAVEFORMATEX *pFormat,
    ppClosestMatch          // [out]  WAVEFORMATEX **ppClosestMatch
  );

  if (NULL!=pClosestMatch) {
#if defined (YASAPI_DEBUG) // {
    if (debug) {
      fputs("  closest match: ",stderr);
      WFXPrint(pClosestMatch,stderr);
      fputs("\n",stderr);
    }
#endif // }
    CoTaskMemFree(pClosestMatch);
    pClosestMatch=NULL;
  }

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_UNSUPPORTED_FORMAT,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGEV(debug,"format not supported (channels: %d, rate: %d, bps: %d"
        ", %s)",pwfx->nChannels,pwfx->nSamplesPerSec,pwfx->wBitsPerSample,
        WFXXGetSubFormat(pwfxx));
  }

#if defined (YASAPI_DEBUG) // {
    if (debug) {
      fputs("  format supported: ",stderr);
      WFXPrint(&pPlayer->wfxx.Format,stderr);
      fputs("\n",stderr);
    }
#endif // }
#endif // }

  // get the device period.
  if (0==hnsDevicePeriod) {
#if 1 // {
    // get the minimum device period (low latency).
    hr=pAudioClient->lpVtbl->GetDevicePeriod(pAudioClient,
      NULL,               // [out]  REFERENCE_TIME *phnsDefaultDevicePeriod,
      &hnsDevicePeriod    // [out]  REFERENCE_TIME *phnsMinimumDevicePeriod
    );
#else // } {
    // get the default device period.
    hr=pAudioClient->lpVtbl->GetDevicePeriod(pAudioClient,
      &hnsDevicePeriod,   // [out]  REFERENCE_TIME *phnsDefaultDevicePeriod,
      NULL                // [out]  REFERENCE_TIME *phnsMinimumDevicePeriod
    );
#endif // }

    if (FAILED(hr)) {
      DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
      DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
      DERROR(debug,E_POINTER,hr);
      DMESSAGE(debug,"getting device period");
      goto period;
    }

    DPRINTF(debug,"  got device period (device period: %d)\n",hnsDevicePeriod);

    if (1.0<qShareSize) {
      hnsDevicePeriod=qShareSize*hnsDevicePeriod+0.5;
      DPRINTF(debug,"  device period scaled by %.02f (device period: %d)\n",
          qShareSize,hnsDevicePeriod);
    }

    // save hnsDevicePeriod into device options.
    OptionsSetDevicePeriod(pOptions,hnsDevicePeriod);
  }
#if ! defined (YASAPI_AUDCLNT_SHAREMODE) // {
retry:
#endif // }
  // set the stream flags.
  dwStreamFlags=0;

  if (AUDCLNT_SHAREMODE_SHARED==eShareMode&&autoconvert_pcm) {
  	  dwStreamFlags|=AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM;

     if (src_default_quality)
  	   dwStreamFlags|=AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY;
  }

  // initialize the audio client.
  hr=pPlayer->strategy->InitializeAudioClient(pPlayer,eShareMode,
      dwStreamFlags,hnsDevicePeriod);

  if (FAILED(hr)) {
#if ! defined (YASAPI_AUDCLNT_SHAREMODE) // {
    if (AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED==hr&&0==retry) {
      DPUTS(debug,"  failed with AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED\n");
      DPUTS(debug,"  trying to get an aligned buffer size\n");
      retry=1;

      // 1. Call IAudioClient::GetBufferSize and receive the
      //    next-highest-aligned buffer size (in frames).
      hr=pAudioClient->lpVtbl->GetBufferSize(pAudioClient,
        &uNumBufferFrames     // [out] UINT32 *pNumBufferFrames
      );

      if (FAILED(hr)) {
        DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
        DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
        DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
        DERROR(debug,E_POINTER,hr);
        DMESSAGE(debug,"getting aligned number of frames");
        goto aligned;
      }

      DPRINTF(debug,"  got aligned number of frames: %u\n",
          uNumBufferFrames);

      // 3. Calculate the aligned buffer size in 100-nansecond units (hns).
      //    The buffer size is (REFERENCE_TIME)((10000.0 * 1000 /
      //    WAVEFORMATEX.nSamplesPerSecond * nFrames) + 0.5). In this
      //    formula, nFrames is the buffer size retrieved by GetBufferSize.
      hnsDevicePeriod=PLAYER_FRAMES_TIMER(uNumBufferFrames,pwfx)+0.5;

      // save hnsDevicePeriod into device options.
      OptionsSetDevicePeriod(pOptions,hnsDevicePeriod);

      if (NULL==(pAudioClient=PlayerActivate(pPlayer)))
        goto client;

      // 5. Call Initialize again on the created audio client and
      //    specify the new buffer size and periodicity.
      goto retry;
    }

    if (AUDCLNT_SHAREMODE_EXCLUSIVE==eShareMode) {
      if (YASAPI_MODE_AUTOMATIC==mode) {
        DPUTS(debug,"  initializing in EXCLUSIVE mode failed\n");
        DPUTS(debug,"  retrying in SHARED mode\n");
        eShareMode=AUDCLNT_SHAREMODE_SHARED;
        retry=0;

        goto retry;
      }
    }
#endif // }

    DERROR(debug,AUDCLNT_E_ALREADY_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(debug,AUDCLNT_E_CPUUSAGE_EXCEEDED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_IN_USE,hr);
    DERROR(debug,AUDCLNT_E_ENDPOINT_CREATE_FAILED,hr);
    DERROR(debug,AUDCLNT_E_INVALID_DEVICE_PERIOD,hr);
    DERROR(debug,AUDCLNT_E_UNSUPPORTED_FORMAT,hr);
    DERROR(debug,AUDCLNT_E_EXCLUSIVE_MODE_NOT_ALLOWED,hr);
    DERROR(debug,AUDCLNT_E_BUFDURATION_PERIOD_NOT_EQUAL,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGEV(debug,"initializing audio client (mode: %s)",
        YASAPI_SHAREMODE_LABEL(eShareMode));
    goto initialize;
  }

  DPRINTF(debug,"  audio client initialized (mode: %s, channels: %d, "
      "rate: %d, bps: %d)\n",
      YASAPI_SHAREMODE_LABEL(eShareMode),pwfx->nChannels,
      pwfx->nSamplesPerSec,pwfx->wBitsPerSample);

#if defined (YASAPI_PLAYER_GET_STREAM_LATENCY) // {
  // get the latency.
  hr=pAudioClient->lpVtbl->GetStreamLatency(pAudioClient,
    &hnsLatency               // [out] REFERENCE_TIME *phnsLatency
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"getting latency");
    goto latency;
  }

  DPUTS(debug,"  got latency\n");
#endif // }

  if (pPlayer->strategy->CreateEvent(pPlayer)<0)
    goto event;

  // get the buffer size.
  //
  // This method retrieves the length of the endpoint buffer shared
  // between the client application and the audio engine. The length
  // is expressed as the number of audio frames the buffer can hold.
  // The size in bytes of an audio frame is calculated as the number
  // of channels in the stream multiplied by the sample size per channel.
  // For example, the frame size is four bytes for a stereo (2-channel)
  // stream with 16-bit samples.
  hr=pAudioClient->lpVtbl->GetBufferSize(pAudioClient,
    &pPlayer->nFrames         // [out]  UINT32 *pNumBufferFrames
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"getting buffer size");
    goto size;
  }

  DPRINTF(debug,"  got buffer size (frames: %d)\n",pPlayer->nFrames);

  // get the render client.
  hr=pAudioClient->lpVtbl->GetService(pAudioClient,
    &IID_IAudioRenderClient,  // [in]   REFIID riid,
    (void**)&pRenderClient    // [out]  void **ppv
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGE(debug,"getting render audio client");
    goto render;
  }

  pPlayer->pRenderClient=pRenderClient;
  DPUTS(debug,"  got render audio client\n");

  // create the sync.
#if defined (YASAPI_DEBUG) // {
  sc.debug=debug;
#endif // }
  sc.pClient=pPlayer;
  sc.end.state=PLAYER_END;
  sc.read.state=PLAYER_READ;
  sc.read.Condition=PlayerCanRead;
  sc.write.state=PLAYER_WRITE;
  sc.write.Condition=PlayerCanWrite;

  if (SyncCreate(&pPlayer->sync,&sc)<0) {
    DMESSAGE(debug,"creating sync");
    goto sync;
  }

  DPUTS(debug,"  sync created\n");

  // create the ring buffer.
  if (1.0<qRingSize) {
    uBufSize=qRingSize*pPlayer->nFrames*pwfx->nBlockAlign+0.5;

    if (0<uBufSize%pwfx->nBlockAlign) {
      uBufSize+=pwfx->nBlockAlign;
      uBufSize-=uBufSize%pwfx->nBlockAlign;
    }
  }
  else
    uBufSize=pPlayer->nFrames*pwfx->nBlockAlign;

  if (RingCreate(&pPlayer->ring,uBufSize,debug)<0) {
    DMESSAGEV(debug,"creating ring buffer (scale %.1f, size: %u)",
        qRingSize,uBufSize);
    goto ring;
  }

  pPlayer->ring.Copy=PlayerCopyMemory;
  pPlayer->ring.pClient=pPlayer;

  DPRINTF(debug,"  ring buffer created (scale %.1f, size: %d)\n",
      (double)uBufSize/(pwfx->nBlockAlign*pPlayer->nFrames)+0.5,uBufSize);

  // get the audio clock.
  hr=pAudioClient->lpVtbl->GetService(pAudioClient,
    &IID_IAudioClock,       // [in]   REFIID riid,
    (void**)&pAudioClock    // [out]  void **ppv
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGE(debug,"getting audio clock");
    goto clock;
  }

  pPlayer->pAudioClock=pAudioClock;
  DPUTS(debug,"  got audio clock\n");

  // create the render audio thread.
  pPlayer->hThread=CreateThread(
    NULL,           // _In_opt_   LPSECURITY_ATTRIBUTES lpThreadAttributes,
    0,              // _In_       SIZE_T dwStackSize,
    PlayerThread,   // _In_       LPTHREAD_START_ROUTINE lpStartAddress,
    pPlayer,        // _In_opt_   LPVOID lpParameter,
    0,              // _In_       DWORD dwCreationFlags,
    NULL            // _Out_opt_  LPDWORD lpThreadId
  );

  if (NULL==pPlayer->hThread) {
    DMESSAGE(debug,"creating render audio thread");
    goto thread;
  }

  DPUTS(debug,"  render audio thread created\n");

  pPlayer->initialized=1;

  return 0;
// cleanup:
thread:
  pAudioClock->lpVtbl->Release(pAudioClock);
clock:
  RingDestroy(&pPlayer->ring);
ring:
  SyncDestroy(&pPlayer->sync);
sync:
  pRenderClient->lpVtbl->Release(pRenderClient);
render:
size:
  pPlayer->strategy->DestroyEvent(pPlayer);
event:
#if defined (YASAPI_PLAYER_GET_STREAM_LATENCY) // {
latency:
#endif // }
aligned:
initialize:
period:
format:
  pAudioClient->lpVtbl->Release(pAudioClient);
client:
initialized:
  return -1;
}

void PlayerDestroy(Player *pPlayer)
{
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  IAudioClock *pAudioClock=pPlayer->pAudioClock;
  IAudioRenderClient *pRenderClient=pPlayer->pRenderClient;

  PlayerShutdown(pPlayer);
  pPlayer->initialized=0;

  CloseHandle(pPlayer->hThread);
  DPUTS(debug,"  render audio thread destroyed\n");

  pAudioClock->lpVtbl->Release(pAudioClock);
  DPUTS(debug,"  render audio clock destroyed\n");

  RingDestroy(&pPlayer->ring);
  DPUTS(debug,"  ring buffer destroyed\n");
  SyncDestroy(&pPlayer->sync);
  DPUTS(debug,"  sync destroyed\n");

  pPlayer->strategy->DestroyEvent(pPlayer);

  pRenderClient->lpVtbl->Release(pRenderClient);
  DPUTS(debug,"  render audio client destroyed\n");

  pAudioClient->lpVtbl->Release(pAudioClient);
  DPUTS(debug,"  audio client destroyed\n");

  pPlayer->pPlugin=NULL;
}

///////////////////////////////////////////////////////////////////////////////
void PlayerShutdown(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  IAudioClock *pAudioClock=pPlayer->pAudioClock;
  Sync *pSync=&pPlayer->sync;

  pPlayer->pAudioClock=NULL;
  WaitForSingleObject(pSync->hMutex,INFINITE);
  pSync->state|=PLAYER_END;
  SetEvent(pSync->read.hEvent);
  SetEvent(pSync->write.hEvent);
  ReleaseMutex(pSync->hMutex);

  WaitForSingleObject(pPlayer->hThread,INFINITE);
  pPlayer->pAudioClock=pAudioClock;
  DPUTS(debug,"  player shut down\n");
}

#if defined (YASAPI_DEBUG) // {
int PlayerGetDebug(Player *pPlayer)
{
  return pPlayer->pPlugin->pOptions->debug;
}
#endif // }

int PlayerGetMaxLatencyMs(Player *pPlayer)
{
  Options *pOptions=pPlayer->pPlugin->pOptions;
  double qRingFill=OptionsGetRingFill(pOptions);
  int nFrames=pPlayer->nFrames;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;

  return qRingFill*PLAYER_FRAMES_MS(nFrames,pwfx)+0.5;
}

HRESULT PlayerStart(Player *pPlayer)
{
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  Sync *pSync=&pPlayer->sync;
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr=S_OK;

  WaitForSingleObject(pSync->hMutex,INFINITE);

  if (0==((PLAYER_STARTED|PLAYER_END)&pSync->state)) {
    // start the audio client.
    hr=pAudioClient->lpVtbl->Start(pAudioClient);

    if (FAILED(hr)) {
      DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
      DERROR(debug,AUDCLNT_E_NOT_STOPPED,hr);
      DERROR(debug,AUDCLNT_E_EVENTHANDLE_NOT_SET,hr);
      DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
      DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
      DMESSAGE(debug,"starting audio client");
      goto start;
    }

    pSync->state|=PLAYER_STARTED;
    DPUTS(debug,"  audio client started\n");
  }
start:
  ReleaseMutex(pSync->hMutex);
  return hr;
}

HRESULT PlayerStopAtom(Player *pPlayer)
{
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  Sync *pSync=&pPlayer->sync;
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr=S_OK;

  if (PLAYER_IS_STARTED(pSync)) {
    // start the audio client.
    hr=pAudioClient->lpVtbl->Stop(pAudioClient);

    if (FAILED(hr)) {
      DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
      DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING ,hr);
      DMESSAGE(debug,"stopping audio client");
      goto start;
    }

    DPUTS(debug,"  audio client stopped\n");
    pSync->state&=~PLAYER_STARTED;
  }
start:
  return hr;
}

HRESULT PlayerStop(Player *pPlayer)
{
  HANDLE hMutex=pPlayer->sync.hMutex;
  HRESULT hr;

  WaitForSingleObject(hMutex,INFINITE);
  hr=PlayerStopAtom(pPlayer);
  ReleaseMutex(hMutex);

  return hr;
}

DWORD PlayerGetTime(Player *pPlayer)
{
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  IAudioClock *pAudioClock=pPlayer->pAudioClock;
  DWORD dwTime=0;
  UINT64 u64Position;
  UINT64 u64Frequency;
  HRESULT hr;

  // not initialized.
  if (NULL==pAudioClock)
    goto clock;

  EnterCriticalSection(&pPlugin->cs);

  // getting the position.
  hr=pAudioClock->lpVtbl->GetPosition(pAudioClock,
    &u64Position,     // [out]  UINT64 *pu64Position,
    NULL              // [out]  UINT64 *pu64QPCPosition
  );

  if (FAILED(hr)) {
    DERROR(debug,S_FALSE,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGE(debug,"getting position");
    goto position;
  }

  // getting the freqency.
  hr=pAudioClock->lpVtbl->GetFrequency(pAudioClock,
    &u64Frequency     // [out]  UINT64 *pu64Frequency
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGE(debug,"getting frequency");
    goto frequency;
  }

  dwTime=1000.0*u64Position/u64Frequency+0.5;
frequency:
position:
  LeaveCriticalSection(&pPlugin->cs);
clock:
  //DPRINTF(debug,"time: %d\n",dwTime);

  return dwTime;
}

WORD PlayerSamplesPerSec(Player *pPlayer)
{
  return pPlayer->wfxx.Format.nSamplesPerSec;
}

WORD PlayerChannels(Player *pPlayer)
{
  return 0==pPlayer->mono?pPlayer->wfxx.Format.nChannels:1;
}

WORD PlayerBlockAlign(Player *pPlayer)
{
  return PlayerChannels(pPlayer)*pPlayer->wBytesPerSample;
}

SIZE_T PlayerAvailable(Player *pPlayer)
{
  return 0==pPlayer->mono
      ?pPlayer->ring.dwAvailable
      :pPlayer->ring.dwAvailable>>1;
}

BOOL PlayerCanWrite(LPVOID pClient)
{
  return 0<((Player *)pClient)->ring.dwAvailable;
}

BOOL PlayerCanRead(LPVOID pClient)
{
  Player *pPlayer=pClient;
  Sync *pSync=&pPlayer->sync;

  return !PLAYER_IS_PAUSE(pSync)&&pPlayer->strategy->CanRead(pPlayer);
}

////////
SIZE_T PlayerWriteWait(Player *pPlayer, LPSTR pData, SIZE_T dwSize)
{
#if defined (YASAPI_DEBUG) // {
  int debug=PlayerGetDebug(pPlayer);
#endif // }
  Sync *pSync=&pPlayer->sync;
  Ring *pRing=&pPlayer->ring;

  dwSize-=dwSize%pPlayer->wfxx.Format.nBlockAlign;

  if (0!=pPlayer->mono)
    dwSize<<=1;

  SyncWaitWrite(pSync);
//DPRINTF(debug,"WRITE (%d, %d) -> ",pRing->dwWritten,pRing->dwAvailable);
  dwSize=RingWrite(pRing,pData,dwSize,0==pPlayer->mono?1:2);
//DPRINTF(debug,"(%d, %d)\n",pRing->dwWritten,pRing->dwAvailable);
  SyncReleaseWrite(pSync);

  if (0!=pPlayer->mono)
    dwSize>>=1;

  return dwSize;
}

SIZE_T PlayerWrite(Player *pPlayer, LPSTR pData, SIZE_T dwSize)
{
  return PlayerWriteWait(pPlayer,pData,dwSize);
}

HRESULT PlayerResetAtom(Player *pPlayer)
{
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr;

  // get reset.
  hr=pAudioClient->lpVtbl->Reset(pAudioClient);

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_NOT_STOPPED,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_OPERATION_PENDING,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGE(debug,"resetting audio client");
    goto reset;
  }
reset:
  return hr;
}

void PlayerReset(Player *pPlayer)
{
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  Sync *pSync=&pPlayer->sync;
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr;

  WaitForSingleObject(pSync->hMutex,INFINITE);

  // stop the player.
  if (PLAYER_IS_STARTED(pSync)) {
    hr=PlayerStopAtom(pPlayer);

    if (FAILED(hr)) {
      DMESSAGE(debug,"stopping audio client");
      goto stop;
    }
  }

  hr=PlayerResetAtom(pPlayer);

  if (FAILED(hr)) {
    DMESSAGE(debug,"resetting audio client");
    goto reset;
  }

  // reset the ring buffer.
  RingReset(&pPlayer->ring);
reset:
stop:
  ReleaseMutex(pSync->hMutex);
}

void PlayerPause(Player *pPlayer, BOOL bPause)
{
  Sync *pSync=&pPlayer->sync;

  WaitForSingleObject(pSync->hMutex,INFINITE);

  if (0==bPause)
    pSync->state&=~PLAYER_PAUSE;
  else {
    pSync->state|=PLAYER_PAUSE;
    PlayerStopAtom(pPlayer);
  }

  SetEvent(pSync->read.hEvent);
  SetEvent(pSync->write.hEvent);
  ReleaseMutex(pSync->hMutex);
}

///////////////////////////////////////////////////////////////////////////////
HRESULT PlayerRenderFramesUnlocked(Player *pPlayer, UINT32 nFramesWrite,
    DWORD dwFlags)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  IAudioRenderClient *pRenderClient=pPlayer->pRenderClient;
  Ring *pRing=&pPlayer->ring;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
  BYTE *pData;
  HRESULT hr;

  hr=pRenderClient->lpVtbl->GetBuffer(pRenderClient,
    nFramesWrite,       // [in]   UINT32 NumFramesRequested,
    &pData              // [out]  BYTE **ppData
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_BUFFER_ERROR,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_TOO_LARGE,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(debug,AUDCLNT_E_OUT_OF_ORDER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_OPERATION_PENDING,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"getting render buffer");
    goto get;
  };


  if (0<nFramesWrite) {
    // transfer the audio data from the ring buffer into the render buffer.
//DPRINTF(debug,"READ (%d, %d) -> ",pRing->dwWritten,pRing->dwAvailable);
    RingRead(pRing,pData,nFramesWrite*pwfx->nBlockAlign);
//DPRINTF(debug,"(%d, %d)\n",pRing->dwWritten,pRing->dwAvailable);
  }

  // release the render buffer.
  hr=pRenderClient->lpVtbl->ReleaseBuffer(pRenderClient,
    nFramesWrite,     // [in]  UINT32 NumFramesWritten,
    dwFlags           // [in]  DWORD dwFlags
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_INVALID_SIZE,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(debug,AUDCLNT_E_OUT_OF_ORDER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DMESSAGE(debug,"releasing render buffer");
    goto release;
  }
release:
get:
  return hr;
}

HRESULT PlayerRenderFrames(Player *pPlayer, UINT32 *pNumFramesRender)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  Sync *pSync=&pPlayer->sync;
  UINT32 nFramesPadding;
  UINT32 nFramesWrite;
  HRESULT hr;

  // lock the ring buffer.
  SyncWaitRead(pSync);

  if (PLAYER_IS_PAUSE(pSync)) {
    // if we are going here and are in "pause" we're in "end" too.
    if (!PLAYER_IS_END(pSync))
      DMESSAGE(debug,"inconsistent state");

    RingReset(&pPlayer->ring);
    *pNumFramesRender=0;
  }
  else {
    hr=pPlayer->strategy->GetFramesWrite(pPlayer,&nFramesWrite,
        &nFramesPadding);

    if (FAILED(hr))
      goto frames;

    // get the render buffer.
    if (0<nFramesWrite) {
      hr=PlayerRenderFramesUnlocked(pPlayer,nFramesWrite,0);

      if (FAILED(hr)) {
        DMESSAGE(debug,"rendering");
        goto render;
      }
    }

    *pNumFramesRender=nFramesPadding+nFramesWrite;
  }
render:
frames:
  // release the lock on the ring buffer.
  SyncReleaseRead(pSync);
  return hr;
}

void PlayerFlush(Player *pPlayer)
{
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  Sync *pSync=&pPlayer->sync;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
  UINT32 nFrames,nFramesRender;
  HRESULT hr;

  PlayerRenderFramesUnlocked(pPlayer,0,AUDCLNT_BUFFERFLAGS_SILENT);

  if (PLAYER_IS_PAUSE(pSync))
    PlayerResetAtom(pPlayer);
  else {
    nFrames=0;

    for (;;) {
      // get number of remaining frames in render buffer.
      hr=pAudioClient->lpVtbl->GetCurrentPadding(pAudioClient,
        &nFramesRender      // [out]  UINT32 *pNumPaddingFrames
      );

      if (FAILED(hr))
        break;
      else if (nFrames==nFramesRender) {
#if defined (YASAPI_DEBUG) // {
        if (0<nFrames)
          DPRINTF(debug, "  sorry, %d frames lost ...\n",nFrames);
#endif // }
        break;
      }

      // remember the numer of frames in order to avoid an endless loop.
      nFrames=nFramesRender;

      // in case there are frames, wait until played.
      TimerSleep(&pPlugin->timer,PLAYER_FRAMES_TIMER(nFrames,pwfx)+0.5);
    }
  }
}

DWORD WINAPI PlayerThread(LPVOID lpParameter)
{
  Player *pPlayer=lpParameter;
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  Options *pOptions=pPlugin->pOptions;
  double qRingFill=OptionsGetRingFill(pOptions);
  Sync *pSync=&pPlayer->sync;
  Ring *pRing=&pPlayer->ring;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
  UINT32 nFrames=pPlayer->nFrames;
  LONGLONG time=0.5*PLAYER_FRAMES_TIMER(nFrames,pwfx)+0.5;
  UINT32 nFramesRender;
  HANDLE hTask;
  DWORD taskIndex;
  HRESULT hr;

  DPUTS(debug,"start playing\n");
  taskIndex=0;
  hTask=AvSetMmThreadCharacteristicsA("Pro Audio",&taskIndex);

  if (NULL==hTask) {
    DMESSAGE(debug,"setting thread cahracteristics");
    goto task;
  }

  if (1.0<qRingFill)
    nFrames=qRingFill*nFrames+0.5;

  while (!PLAYER_IS_END(pSync)) {
    if (PLAYER_IS_STARTED(pSync))
      pPlayer->strategy->Wait(pPlayer,time);

    hr=PlayerRenderFrames(pPlayer,&nFramesRender);

    if (FAILED(hr)) {
      DMESSAGE(debug,"rendering frames");
      continue;
    }

    // make sure that the renderer can be completely filled before starting.
    if (!PLAYER_IS_STARTED(pSync)&&nFrames<=pRing->dwWritten)
      PlayerStart(pPlayer);
  }

  if (PLAYER_IS_STARTED(pSync)) {
    PlayerFlush(pPlayer);
    PlayerStop(pPlayer);
  }

  DPUTS(debug,"  stop playing\n");
// cleanup:
  AvRevertMmThreadCharacteristics(hTask);
task:
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
void PlayerCopyMemory(PVOID pClient, PVOID Destination, const VOID *Source,
    SIZE_T Length)
{
  Player *pPlayer=pClient;
  Plugin *pPlugin=pPlayer->pPlugin;
  Options *pOptions=pPlugin->pOptions;
  double volume=pPlugin->volume;
  char *wp,*mp;
  const char *rp;
  int m,n;
  double q;
  int32_t i32;

  wp=Destination;
  mp=wp+Length;
  rp=Source;
  n=pPlayer->wfxx.Format.wBitsPerSample;
  m=n>>3;
  q=volume*YASAPI_INT_MAX(n)/YASAPI_INT_MAX(32);

  n=32-n;

  if (0==pPlayer->mono) {
    if (0==pOptions->volume||1.0==volume) {
      CopyMemory(
        Destination,            // _In_  PVOID Destination,
        Source,                 // _In_  const VOID *Source,
        Length                  // _In_  SIZE_T Length
      );
    }
    else {
      while (wp<mp) {
        CopyMemory(
          &i32,                 // _In_  PVOID Destination,
          rp,                   // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        i32<<=n;
        i32=q*i32+0.5;

        CopyMemory(
          wp,                   // _In_  PVOID Destination,
          &i32,                 // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        wp+=m;
        rp+=m;
      }
    }
  }
  else {
    while (wp<mp) {
      CopyMemory(
        wp,                     // _In_  PVOID Destination,
        rp,                     // _In_  const VOID *Source,
        m                       // _In_  SIZE_T Length
      );

      wp+=m;

      CopyMemory(
        wp,                     // _In_  PVOID Destination,
        rp,                     // _In_  const VOID *Source,
        m                       // _In_  SIZE_T Length
      );

      wp+=m;
      rp+=m;
    }
  }
}
