/*
 * yasapi_plugin.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#if ! defined (PKEY_DeviceInterface_FriendlyName) // {
#include <FunctionDiscoveryKeys_devpkey.h>
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_CONFIG2) // {
static int PluginGetDefaultID(Plugin *pPlugin)
{
  int code=-1;
  Options *pOptions=pPlugin->pOptions;
#if defined (YASAPI_DEBUG) // {
  int debug=pOptions->debug;
#endif // }
  IMMDeviceEnumerator *pEnumerator=pPlugin->pEnumerator;
  IMMDevice *pDevice;
  LPWSTR pstrId;
  HRESULT hr;

  // get the default audio device.
  hr=pEnumerator->lpVtbl->GetDefaultAudioEndpoint(pEnumerator,
    eRender,                  // [in]   EDataFlow dataFlow,
    eMultimedia,              // [in]   ERole role,
    &pDevice                  // [out]  IMMDevice **ppDevice
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_NOTFOUND,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGE(debug,"getting default audio device");
    goto device;
  }

  // retrieving an endpoint ID string that identifies the
  // audio endpoint device.
  hr=pDevice->lpVtbl->GetId(pDevice,
    &pstrId                 // [out] LPWSTR *ppstrId
  );

  if (FAILED(hr)) {
    DERROR(debug,E_OUTOFMEMORY,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"retrieving endpoint ID string");
    goto strid;
  }

  wcsncpy(pOptions->id,pstrId,YASAPI_ID_SIZE-1);
  pOptions->id[YASAPI_ID_SIZE-1]=L'\0';
  code=0;

// cleanup:
  CoTaskMemFree(pstrId);
strid:
  pDevice->lpVtbl->Release(pDevice);
device:
  return code;
}
#endif // }

int PluginCreate(Plugin *pPlugin, Options *pOptions)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pOptions->debug;
#else // } {
  int debug=0;
#endif // }
  IMMDeviceEnumerator *pEnumerator;
#if defined (YASAPI_CONFIG2) // {
  IMMDeviceCollection *pCollection;
  UINT cDevices,nDevice;
#else // } {
  IMMDevice *pDevice;
#endif // }
  HRESULT hr;

  // create the critical section.
  InitializeCriticalSection(&pPlugin->cs);
  DPUTS(debug,"  critical section created\n");

  pPlugin->pOptions=pOptions;
  pPlugin->volume=1.0;

  // create the timer.
  if (TimerCreate(&pPlugin->timer,debug)<0) {
    DMESSAGE(debug,"creating timer");
    goto timer;
  }

  DPUTS(debug,"  waitable timer created\n");

  // create a device enumerator.
  hr=CoCreateInstance(
    &CLSID_MMDeviceEnumerator,  // _In_   REFCLSID rclsid,
    NULL,                       // _In_   LPUNKNOWN pUnkOuter,
    CLSCTX_ALL,                 // _In_   DWORD dwClsContext,
    &IID_IMMDeviceEnumerator,   // _In_   REFIID riid,
    (void**)&pEnumerator        // _Out_  LPVOID *ppv
  );

  if (FAILED(hr)) {
    DERROR(debug,REGDB_E_CLASSNOTREG,hr);
    DERROR(debug,CLASS_E_NOAGGREGATION,hr);
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"creating device enumerator");
    goto devnum;
  }

  pPlugin->pEnumerator=pEnumerator;
  DPUTS(debug,"  device enumerator created\n");

#if defined (YASAPI_CONFIG2) // {
  if (L'\0'==*pOptions->id&&PluginGetDefaultID(pPlugin)<0) {
    DMESSAGE(debug,"getting default id");
    goto id1;
  }

  // get the audio device collection.
  hr=pEnumerator->lpVtbl->EnumAudioEndpoints(pEnumerator,
    eRender,                  // [in]  EDataFlow           dataFlow,
    DEVICE_STATE_ACTIVE,      // [in]  DWORD               dwStateMask,
    &pCollection              // [out] IMMDeviceCollection **ppDevices
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGE(debug,"creating device collection");
    goto collection;
  }

  pPlugin->pCollection=pCollection;
  DPUTS(debug,"  device collection created\n");

  // get the number of audio devices.
  hr=pCollection->lpVtbl->GetCount(pCollection,
    &cDevices                 // [out] UINT *pcDevices
  );

  if (FAILED(hr)) {
    DMESSAGE(debug,"getting the number of audio devices");
    goto cdevices;
  }

  DPRINTF(debug,"  got number of audio devices: %u\n",cDevices);
  pOptions->cDevices=0;

  // allocate device options.
  pOptions->pDevices=calloc(cDevices,sizeof *pOptions->pDevices);

  if (NULL==pOptions->pDevices) {
    DMESSAGE(debug,"allocating device options");
    goto calloc;
  }

  for (nDevice=0;nDevice<cDevices;++nDevice) {
    DPRINTF(debug,"  device %d:\n",nDevice+1);

    if (DeviceOptionsInit(pOptions->pDevices+nDevice,pPlugin,nDevice)<0) {
      DMESSAGE(debug,"initializing device options");
      goto loop;
    }

    if (0==wcscmp(pOptions->id,pOptions->pDevices[nDevice].id))
      pOptions->nDevice=nDevice;

    ++pOptions->cDevices;
  }

  if (pOptions->nDevice<0) {
    if (PluginGetDefaultID(pPlugin)<0) {
      DMESSAGE(debug,"getting default id");
      goto id2;
    }

    for (nDevice=0;nDevice<cDevices;++nDevice) {
      if (0==wcscmp(pOptions->id,pOptions->pDevices[nDevice].id)) {
        OptionsSetDevice(pOptions,nDevice);
        break;
      }
    }

    if (pOptions->nDevice<0) {
      DMESSAGE(debug,"missing default device");
      goto id3;
    }
  }
#else // } {
  // get the default audio device.
  hr=pEnumerator->lpVtbl->GetDefaultAudioEndpoint(pEnumerator,
    eRender,                  // [in]   EDataFlow dataFlow,
    eMultimedia,              // [in]   ERole role,
    &pDevice                  // [out]  IMMDevice **ppDevice
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_NOTFOUND,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGE(debug,"getting default audio device");
    goto device;
  }

  pPlugin->pDevice=pDevice;
  DPUTS(debug,"  device created\n");
#endif // }

  // yep!
  pPlugin->initialized=1;

  return 0;
// cleanup:
#if ! defined (YASAPI_CONFIG2) // {
  pDevice->lpVtbl->Release(pDevice);
device:
#else // {
id3:
id2:
loop:
  while (0<pOptions->cDevices) {
    --pOptions->cDevices;
    DeviceOptionsCleanup(pOptions->pDevices+pOptions->cDevices);
  }

  free(pOptions->pDevices);
calloc:
cdevices:
  pCollection->lpVtbl->Release(pCollection);
collection:
id1:
#endif // }
  pEnumerator->lpVtbl->Release(pEnumerator);
devnum:
  TimerDestroy(&pPlugin->timer);
timer:
  DeleteCriticalSection(&pPlugin->cs);
  return -1;
}

void PluginDestroy(Plugin *pPlugin)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  IMMDeviceEnumerator *pEnumerator=pPlugin->pEnumerator;
#if defined (YASAPI_CONFIG2) // {
  Options *pOptions=pPlugin->pOptions;
  IMMDeviceCollection *pCollection=pPlugin->pCollection;
#else // } {
  IMMDevice *pDevice=pPlugin->pDevice;
#endif // }

  pPlugin->initialized=0;

#if ! defined (YASAPI_CONFIG2) // {
  pDevice->lpVtbl->Release(pDevice);
  DPUTS(debug,"  device destroyed\n");
#else // {
  while (0<pOptions->cDevices) {
    --pOptions->cDevices;
    DeviceOptionsCleanup(pOptions->pDevices+pOptions->cDevices);
  }

  free(pOptions->pDevices);
  pCollection->lpVtbl->Release(pCollection);
  DPUTS(debug,"  device collection destroyed\n");
#endif // }
  pEnumerator->lpVtbl->Release(pEnumerator);
  DPUTS(debug,"  device enumerator destroyed\n");
  TimerDestroy(&pPlugin->timer);
  DPUTS(debug,"  waitable timer destroyed\n");

  DeleteCriticalSection(&pPlugin->cs);
  DPUTS(debug,"  critical section destroyed\n");
}
